import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class Shard(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        hostname (str): hostname
        locales (list<str>): locales
        name (str): name
        region_tag (str): region tag
        slug (str): slug
    """
    def __init__(self, dictionary):
        self.hostname = dictionary.get("hostname", "")
        self.locales = dictionary.get("locales", [])
        self.name = dictionary.get("name", "")
        self.region_tag = dictionary.get("region_tag", "")
        self.slug = dictionary.get("slug", "")


@cassiopeia.type.core.common.inheritdocs
class Translation(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        content (str): content
        locale (str): locale
        updated_at (str): timestamp
    """
    def __init__(self, dictionary):
        self.content = dictionary.get("content", "")
        self.locale = dictionary.get("locale", "")
        self.updated_at = dictionary.get("updated_at", "")


@cassiopeia.type.core.common.inheritdocs
class Message(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        author (str): author
        content (str): content
        created_at (str): timestamp created
        id (int): ID
        severity (str): legal values: Info, Alert, Error
        translations (list<Translation>): translations
        updated_at (str): timestamp updated
    """
    def __init__(self, dictionary):
        self.author = dictionary.get("author", "")
        self.content = dictionary.get("content", "")
        self.created_at = dictionary.get("created_at", "")
        self.id = dictionary.get("id", 0)
        self.severity = dictionary.get("severity", "")
        self.translations = [(Translation(trans) if not isinstance(trans, Translation) else trans) for trans in dictionary.get("translations", []) if trans]
        self.updated_at = dictionary.get("updated_at", "")


@cassiopeia.type.core.common.inheritdocs
class Incident(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        active (bool): active
        created_at (str): timestamp created
        id (int): ID
        updates (list<Message>): updates
    """
    def __init__(self, dictionary):
        self.active = dictionary.get("active", False)
        self.created_at = dictionary.get("created_at", "")
        self.id = dictionary.get("id", 0)
        self.updates = [(Message(msg) if not isinstance(msg, Message) else msg) for msg in dictionary.get("updates", []) if msg]


@cassiopeia.type.core.common.inheritdocs
class Service(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        incidents (list<Incident>): incidents
        name (str): name
        slug (str): slug
        status (str): legal values: Online, Alert, Offline, Deploying
    """
    def __init__(self, dictionary):
        self.incidents = [(Incident(inc) if not isinstance(inc, Incident) else inc) for inc in dictionary.get("incidents", []) if inc]
        self.name = dictionary.get("name", "")
        self.slug = dictionary.get("slug", "")
        self.status = dictionary.get("status", "")


@cassiopeia.type.core.common.inheritdocs
class ShardStatus(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        hostname (str): hostname
        locales (list<string>): locales
        name (str): name
        region_tag (str): region tag
        services (list<Service>): services
        slug (str): slug
    """
    def __init__(self, dictionary):
        self.hostname = dictionary.get("hostname", "")
        self.locales = dictionary.get("locales", [])
        self.name = dictionary.get("name", "")
        self.region_tag = dictionary.get("region_tag", "")
        self.services = [(Service(srvc) if not isinstance(srvc, Service) else srvc) for srvc in dictionary.get("services", []) if srvc]
        self.slug = dictionary.get("slug", "")


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_shard():
    global Shard

    @cassiopeia.type.core.common.inheritdocs
    class Shard(Shard, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Shard"
        hostname = sqlalchemy.Column(sqlalchemy.String(50))
        locales = sqlalchemy.Column(cassiopeia.type.dto.common.JSONEncoded)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        region_tag = sqlalchemy.Column(sqlalchemy.String(30), primary_key=True)
        slug = sqlalchemy.Column(sqlalchemy.String(30))


def _sa_bind_translation():
    global Translation

    @cassiopeia.type.core.common.inheritdocs
    class Translation(Translation, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Translation"
        content = sqlalchemy.Column(sqlalchemy.Text)
        locale = sqlalchemy.Column(sqlalchemy.String(30))
        updated_at = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _message_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("Message.id", ondelete="CASCADE"))


def _sa_bind_message():
    global Message

    @cassiopeia.type.core.common.inheritdocs
    class Message(Message, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Message"
        author = sqlalchemy.Column(sqlalchemy.String(30))
        content = sqlalchemy.Column(sqlalchemy.Text)
        created_at = sqlalchemy.Column(sqlalchemy.String(30))
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        severity = sqlalchemy.Column(sqlalchemy.String(30))
        translations = sqlalchemy.orm.relationship("cassiopeia.type.dto.status.Translation", cascade="all, delete-orphan", passive_deletes=True)
        updated_at = sqlalchemy.Column(sqlalchemy.String(30))
        _incident_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("Incident.id", ondelete="CASCADE"))


def _sa_bind_incident():
    global Incident

    @cassiopeia.type.core.common.inheritdocs
    class Incident(Incident, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Incident"
        active = sqlalchemy.Column(sqlalchemy.Boolean)
        created_at = sqlalchemy.Column(sqlalchemy.String(30))
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        updates = sqlalchemy.orm.relationship("cassiopeia.type.dto.status.Message", cascade="all, delete-orphan", passive_deletes=True)
        _service_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("Service._id", ondelete="CASCADE"))


def _sa_bind_service():
    global Service

    @cassiopeia.type.core.common.inheritdocs
    class Service(Service, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Service"
        incidents = sqlalchemy.orm.relationship("cassiopeia.type.dto.status.Incident", cascade="all, delete-orphan", passive_deletes=True)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        slug = sqlalchemy.Column(sqlalchemy.String(30))
        status = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _shard_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("ShardStatus._id", ondelete="CASCADE"))


def _sa_bind_shard_status():
    global ShardStatus

    @cassiopeia.type.core.common.inheritdocs
    class ShardStatus(ShardStatus, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "ShardStatus"
        hostname = sqlalchemy.Column(sqlalchemy.String(50))
        locales = sqlalchemy.Column(cassiopeia.type.dto.common.JSONEncoded)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        region_tag = sqlalchemy.Column(sqlalchemy.String(30))
        services = sqlalchemy.orm.relationship("cassiopeia.type.dto.status.Service", cascade="all, delete-orphan", passive_deletes=True)
        slug = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)


def _sa_bind_all():
    _sa_bind_shard()
    _sa_bind_translation()
    _sa_bind_message()
    _sa_bind_incident()
    _sa_bind_service()
    _sa_bind_shard_status()



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/gameapi tests...")
    test_recent_games()


def test_recent_games():
    int_test_handler.test_result(riotapi.get_recent_games(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))



import cassiopeia.dto.requests
import cassiopeia.type.dto.team


def get_teams_by_summoner_id(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/986/3358

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to look up teams for

    Returns:
        dict<str, list<Team>>: the requested summoners' teams
    """
    # Can only have 10 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 10:
        raise ValueError("Can only get up to 10 summoners' teams at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/team/by-summoner/{ids}".format(version=cassiopeia.dto.requests.api_versions["team"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, teams in response.items():
        response[id_] = [cassiopeia.type.dto.team.Team(team) for team in teams]

    return response


def get_teams_by_id(team_ids):
    """
    https://developer.riotgames.com/api/methods#!/986/3358

    Args:
        team_ids (str | list<str>): the team ID(s) to look up

    Returns:
        dict<str, Team>: the requested teams
    """
    # Can only have 10 teams max if it's a list
    if isinstance(team_ids, list) and len(team_ids) > 10:
        raise ValueError("Can only get up to 10 teams at once.")

    id_string = ",".join(team_ids) if isinstance(team_ids, list) else team_ids

    # Get JSON response
    request = "{version}/team/{ids}".format(version=cassiopeia.dto.requests.api_versions["team"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, team in response.items():
        response[id_] = cassiopeia.type.dto.team.Team(team)

    return response



import cassiopeia.dto.featuredgamesapi
import cassiopeia.core.requests
import cassiopeia.type.core.featuredgames


def get_featured_games():
    """
    Gets the current featured game list

    Returns:
        list<Game>: the featured games
    """
    games = cassiopeia.dto.featuredgamesapi.get_featured_games()

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_champions() if games.champion_ids else None
        cassiopeia.riotapi.get_summoner_spells() if games.summoner_spell_ids else None

    return [cassiopeia.type.core.featuredgames.Game(game) for game in games.gameList]



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class Team(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        createDate (int): date that team was created specified as epoch milliseconds
        fullId (str): fullId
        lastGameDate (int): date that last game played by team ended specified as epoch milliseconds
        lastJoinDate (int): date that last member joined specified as epoch milliseconds
        lastJoinedRankedTeamQueueDate (int): date that team last joined the ranked team queue specified as epoch milliseconds
        matchHistory (list<MatchHistorySummary>): matchHistory
        modifyDate (int): date that team was last modified specified as epoch milliseconds
        name (str): name
        roster (Roster): roster
        secondLastJoinDate (int): date that second to last member joined specified as epoch milliseconds
        status (str): status
        tag (str): tag
        teamStatDetails (list<TeamStatDetail>): stat details
        thirdLastJoinDate (int): date that third to last member joined specified as epoch milliseconds
    """
    def __init__(self, dictionary):
        self.createDate = dictionary.get("createDate", 0)
        self.fullId = dictionary.get("fullId", "")
        self.lastGameDate = dictionary.get("lastGameDate", 0)
        self.lastJoinDate = dictionary.get("lastJoinDate", 0)
        self.lastJoinedRankedTeamQueueDate = dictionary.get("lastJoinedRankedTeamQueueDate", 0)
        self.matchHistory = [(MatchHistorySummary(mh) if not isinstance(mh, MatchHistorySummary) else mh) for mh in dictionary.get("matchHistory", []) if mh]
        self.modifyDate = dictionary.get("modifyDate", 0)
        self.name = dictionary.get("name", "")
        val = dictionary.get("roster", None)
        self.roster = Roster(val) if val and not isinstance(val, Roster) else val
        self.secondLastJoinDate = dictionary.get("secondLastJoinDate", 0)
        self.status = dictionary.get("status", "")
        self.tag = dictionary.get("tag", "")
        self.teamStatDetails = [(TeamStatDetail(ts) if not isinstance(ts, TeamStatDetail) else ts) for ts in dictionary.get("teamStatDetails", []) if ts]
        self.thirdLastJoinDate = dictionary.get("thirdLastJoinDate", 0)

    @property
    def summoner_ids(self):
        """
        Gets all summoner IDs contained in this object
        """
        ids = set()
        ids.add(self.roster.ownerId)
        for member in self.roster.memberList:
            ids.add(member.playerId)
        return ids


@cassiopeia.type.core.common.inheritdocs
class MatchHistorySummary(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all summoner IDs contained in this object
    """
    def __init__(self, dictionary):
        self.assists = dictionary.get("assists", 0)
        self.date = dictionary.get("date", 0)
        self.deaths = dictionary.get("deaths", 0)
        self.gameId = dictionary.get("gameId", 0)
        self.gameMode = dictionary.get("gameMode", "")
        self.invalid = dictionary.get("invalid", False)
        self.kills = dictionary.get("kills", 0)
        self.mapId = dictionary.get("mapId", 0)
        self.opposingTeamKills = dictionary.get("opposingTeamKills", 0)
        self.opposingTeamName = dictionary.get("opposingTeamName", "")
        self.win = dictionary.get("win", False)


@cassiopeia.type.core.common.inheritdocs
class Roster(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        assists (int): assists
        date (int): date that match was completed specified as epoch milliseconds
        deaths (int): deaths
        gameId (int): gameId
        gameMode (str): gameMode
        invalid (bool): invalid
        kills (int): kills
        mapId (int): mapId
        opposingTeamKills (int): opposingTeamKills
        opposingTeamName (str): opposingTeamName
        win (bool): win
    """
    def __init__(self, dictionary):
        self.memberList = [(TeamMemberInfo(ts) if not isinstance(ts, TeamMemberInfo) else ts) for ts in dictionary.get("memberList", []) if ts]
        self.ownerId = dictionary.get("ownerId", 0)


@cassiopeia.type.core.common.inheritdocs
class TeamStatDetail(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        memberList (list<TeamMemberInfo>): memberList
        ownerId (int): ownerId
    """
    def __init__(self, dictionary):
        self.averageGamesPlayed = dictionary.get("averageGamesPlayed", 0)
        self.losses = dictionary.get("losses", 0)
        self.teamStatType = dictionary.get("teamStatType", "")
        self.wins = dictionary.get("wins", 0)


@cassiopeia.type.core.common.inheritdocs
class TeamMemberInfo(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        averageGamesPlayed (int): averageGamesPlayed
        losses (int): losses
        teamStatType (str): teamStatType
        wins (int): wins
    """
    def __init__(self, dictionary):
        self.inviteDate = dictionary.get("inviteDate", 0)
        self.joinDate = dictionary.get("joinDate", 0)
        self.playerId = dictionary.get("playerId", 0)
        self.status = dictionary.get("status", "")


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_team():
    global Team

    @cassiopeia.type.core.common.inheritdocs
    class Team(Team, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Team"
        createDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        fullId = sqlalchemy.Column(sqlalchemy.String(50), primary_key=True)
        lastGameDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        lastJoinDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        lastJoinedRankedTeamQueueDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        matchHistory = sqlalchemy.orm.relationship("cassiopeia.type.dto.team.MatchHistorySummary", cascade="all, delete-orphan", passive_deletes=True)
        modifyDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        roster = sqlalchemy.orm.relationship("cassiopeia.type.dto.team.Roster", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        secondLastJoinDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        status = sqlalchemy.Column(sqlalchemy.String(30))
        tag = sqlalchemy.Column(sqlalchemy.String(30))
        teamStatDetails = sqlalchemy.orm.relationship("cassiopeia.type.dto.team.TeamStatDetail", cascade="all, delete-orphan", passive_deletes=True)
        thirdLastJoinDate = sqlalchemy.Column(sqlalchemy.BigInteger)


def _sa_bind_match_history_summary():
    global MatchHistorySummary

    @cassiopeia.type.core.common.inheritdocs
    class MatchHistorySummary(MatchHistorySummary, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "TeamMatchHistorySummary"
        assists = sqlalchemy.Column(sqlalchemy.Integer)
        date = sqlalchemy.Column(sqlalchemy.BigInteger)
        deaths = sqlalchemy.Column(sqlalchemy.Integer)
        gameId = sqlalchemy.Column(sqlalchemy.Integer)
        gameMode = sqlalchemy.Column(sqlalchemy.String(30))
        invalid = sqlalchemy.Column(sqlalchemy.Boolean)
        kills = sqlalchemy.Column(sqlalchemy.Integer)
        mapId = sqlalchemy.Column(sqlalchemy.Integer)
        opposingTeamKills = sqlalchemy.Column(sqlalchemy.Integer)
        opposingTeamName = sqlalchemy.Column(sqlalchemy.String(30))
        win = sqlalchemy.Column(sqlalchemy.Boolean)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _team_id = sqlalchemy.Column(sqlalchemy.String(50), sqlalchemy.ForeignKey("Team.fullId", ondelete="CASCADE"))


def _sa_bind_roster():
    global Roster

    @cassiopeia.type.core.common.inheritdocs
    class Roster(Roster, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Roster"
        memberList = sqlalchemy.orm.relationship("cassiopeia.type.dto.team.TeamMemberInfo", cascade="all, delete-orphan", passive_deletes=True)
        ownerId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _team_id = sqlalchemy.Column(sqlalchemy.String(50), sqlalchemy.ForeignKey("Team.fullId", ondelete="CASCADE"))


def _sa_bind_team_stat_detail():
    global TeamStatDetail

    @cassiopeia.type.core.common.inheritdocs
    class TeamStatDetail(TeamStatDetail, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "TeamStatDetail"
        averageGamesPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        losses = sqlalchemy.Column(sqlalchemy.Integer)
        teamStatType = sqlalchemy.Column(sqlalchemy.String(30))
        wins = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _team_id = sqlalchemy.Column(sqlalchemy.String(50), sqlalchemy.ForeignKey("Team.fullId", ondelete="CASCADE"))


def _sa_bind_team_member_info():
    global TeamMemberInfo

    @cassiopeia.type.core.common.inheritdocs
    class TeamMemberInfo(TeamMemberInfo, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "TeamMemberInfo"
        inviteDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        joinDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        playerId = sqlalchemy.Column(sqlalchemy.Integer)
        status = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _roster_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("Roster._id", ondelete="CASCADE"))


def _sa_bind_all():
    _sa_bind_team()
    _sa_bind_match_history_summary()
    _sa_bind_roster()
    _sa_bind_team_stat_detail()
    _sa_bind_team_member_info()



import cassiopeia.dto.currentgameapi
import cassiopeia.core.requests
import cassiopeia.type.core.currentgame


def get_current_game(summoner):
    """
    Gets the game a summoner is currently in, if they're in one

    Args:
        summoner (Summoner): the summoner to find an active game for

    Returns:
        Game: the game they're in (or None if they aren't in one)
    """
    game = cassiopeia.dto.currentgameapi.get_current_game(summoner.id)

    # Load required data if loading policy is eager
    if game and cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_champions() if game.champion_ids else None
        cassiopeia.riotapi.get_masteries() if game.mastery_ids else None
        cassiopeia.riotapi.get_runes() if game.rune_ids else None
        summoner_ids = game.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_summoner_spells() if game.summoner_spell_ids else None

    return cassiopeia.type.core.currentgame.Game(game) if game else None



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/team tests...")
    test_teams_by_summoner()
    test_teams()


def test_teams_by_summoner():
    int_test_handler.test_result(riotapi.get_teams_by_summoner(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))


def test_teams():
    int_test_handler.test_result(riotapi.get_teams([int_test_handler.team_id]))



import cassiopeia.dto.requests
import cassiopeia.type.dto.stats


def get_ranked_stats(summoner_id, season=None):
    """
    https://developer.riotgames.com/api/methods#!/1018/3452

    Args:
        summoner_id (int): the summoner to get ranked stats for
        season (str): the season to get ranked stats for ("SEASON2015", "SEASON2014", "SEASON3") (default None)

    Returns:
        RankedStats: the ranked stats for the summoner and season specified
    """
    request = "{version}/stats/by-summoner/{id_}/ranked".format(version=cassiopeia.dto.requests.api_versions["stats"], id_=summoner_id)

    params = {}
    if season:
        params["season"] = season

    return cassiopeia.type.dto.stats.RankedStats(cassiopeia.dto.requests.get(request, params))


def get_stats(summoner_id, season=None):
    """
    https://developer.riotgames.com/api/methods#!/1018/3453

    Args:
        summoner_id (int): the summoner to get ranked stats for
        season (str): the season to get ranked stats for ("SEASON2015", "SEASON2014", "SEASON3") (default None)

    Returns:
        PlayerStatsSummaryList: the ranked stats for the summoner and season specified
    """
    request = "{version}/stats/by-summoner/{id_}/summary".format(version=cassiopeia.dto.requests.api_versions["stats"], id_=summoner_id)

    params = {}
    if season:
        params["season"] = season

    return cassiopeia.type.dto.stats.PlayerStatsSummaryList(cassiopeia.dto.requests.get(request, params))



import cassiopeia.riotapi
import cassiopeia.core.requests
import cassiopeia.dto.championmasteryapi
import cassiopeia.type.core.championmastery
import cassiopeia.type.core.common


def get_champion_mastery(summoner, champion):
    """
    Gets the ChampionMastery object for the specified summoner and champion

    Args:
        summoner (Summoner): the summoner to get champion mastery for
        champion (Champion): the desired champion

    Returns:
        ChampionMastery: the summoner's champion mastery value for the specified champion
    """
    champion_mastery = cassiopeia.dto.championmasteryapi.get_champion_mastery(summoner.id, champion.id)
    return cassiopeia.type.core.championmastery.ChampionMastery(champion_mastery)


def get_champion_masteries(summoner):
    """
    Gets all the ChampionMastery objects for the specified summoner

    Args:
        summoner (Summoner): the summoner to get champion mastery for

    Returns:
        dict<Champion, ChampionMastery>: the summoner's champion masteries
    """
    champion_masteries = cassiopeia.dto.championmasteryapi.get_champion_masteries(summoner.id)

    # Always load champions since we'll be using them here
    cassiopeia.riotapi.get_champions()
    return {cassiopeia.riotapi.get_champion_by_id(cm.championId): cassiopeia.type.core.championmastery.ChampionMastery(cm) for cm in champion_masteries}


def get_champion_mastery_score(summoner):
    """
    Gets the total champion mastery score for the specified summoner

    Args:
        summoner (Summoner): the summoner to get champion mastery for

    Returns:
        int: the summoner's total champion mastery score
    """
    return cassiopeia.dto.championmasteryapi.get_champion_mastery_score(summoner.id)


def get_top_champion_masteries(summoner, max_entries=3):
    """
    Gets the top ChampionMastery objects for the specified summoner

    Args:
        summoner (Summoner): the summoner to get champion mastery for
        max_entries (int): the maximum number of entires to retrieve (default 3)

    Returns:
        list<ChampionMastery>: the summoner's top champion masteries
    """
    champion_masteries = cassiopeia.dto.championmasteryapi.get_top_champion_masteries(summoner.id, max_entries)

    # Load required data if loading policy is eager
    if champion_masteries and cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_champions()

    return [cassiopeia.type.core.championmastery.ChampionMastery(cm) for cm in champion_masteries]



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class MiniSeries(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        losses (int): number of current losses in the mini series
        progress (str): string showing the current, sequential mini series progress where 'W' represents a win, 'L' represents a loss, and 'N' represents a game that hasn't been played yet
        target (int): number of wins required for promotion
        wins (int): number of current wins in the mini series
    """
    def __init__(self, dictionary):
        self.losses = dictionary.get("losses", 0)
        self.progress = dictionary.get("progress", "")
        self.target = dictionary.get("target", 0)
        self.wins = dictionary.get("wins", 0)


@cassiopeia.type.core.common.inheritdocs
class LeagueEntry(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        division (str): the league division of the participant
        isFreshBlood (bool): specifies if the participant is fresh blood
        isHotStreak (bool): specifies if the participant is on a hot streak
        isInactive (bool): specifies if the participant is inactive
        isVeteran (bool): specifies if the participant is a veteran
        leaguePoints (int): the league points of the participant
        losses (int): the number of losses for the participant
        miniSeries (MiniSeries): mini series data for the participant. Only present if the participant is currently in a mini series.
        playerOrTeamId (str): the ID of the participant (i.e., summoner or team) represented by this entry
        playerOrTeamName (str): the name of the the participant (i.e., summoner or team) represented by this entry
        wins (int): the number of wins for the participant
    """
    def __init__(self, dictionary):
        self.division = dictionary.get("division", "")
        self.isFreshBlood = dictionary.get("isFreshBlood", False)
        self.isHotStreak = dictionary.get("isHotStreak", False)
        self.isInactive = dictionary.get("isInactive", False)
        self.isVeteran = dictionary.get("isVeteran", False)
        self.leaguePoints = dictionary.get("leaguePoints", 0)
        self.losses = dictionary.get("losses", 0)
        val = dictionary.get("miniSeries", None)
        self.miniSeries = MiniSeries(val) if val and not isinstance(val, MiniSeries) else val
        self.playerOrTeamId = dictionary.get("playerOrTeamId", "")
        self.playerOrTeamName = dictionary.get("playerOrTeamName", "")
        self.wins = dictionary.get("wins", 0)


@cassiopeia.type.core.common.inheritdocs
class League(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        entries (list<LeagueEntry>): the requested league entries
        name (str): this name is an internal place-holder name only. Display and localization of names in the game client are handled client-side.
        participantId (str): specifies the relevant participant that is a member of this league (i.e., a requested summoner ID, a requested team ID, or the ID of a team to which one of the requested summoners belongs). Only present when full league is requested so that participant's entry can be identified. Not present when individual entry is requested.
        queue (str): the league's queue type (Legal values: RANKED_SOLO_5x5, RANKED_TEAM_3x3, RANKED_TEAM_5x5)
        tier (str): the league's tier (Legal values: CHALLENGER, MASTER, DIAMOND, PLATINUM, GOLD, SILVER, BRONZE)
    """
    def __init__(self, dictionary):
        self.entries = [(LeagueEntry(entry) if not isinstance(entry, LeagueEntry) else entry) for entry in dictionary.get("entries", []) if entry]
        self.name = dictionary.get("name", "")
        self.participantId = dictionary.get("participantId", "")
        self.queue = dictionary.get("queue", "")
        self.tier = dictionary.get("tier", "")

    @property
    def summoner_ids(self):
        """
        Gets all summoner IDs contained in this object
        """
        ids = set()

        if self.participantId:
            try:
                id_ = int(self.participantId)
                ids.add(id_)
            except ValueError:
                pass

        for entry in self.entries:
            if entry.playerOrTeamId:
                try:
                    id_ = int(entry.playerOrTeamId)
                    ids.add(id_)
                except ValueError:
                    pass
        return ids

    @property
    def team_ids(self):
        """
        Gets all summoner IDs contained in this object
        """
        ids = set()

        if self.participantId:
            try:
                int(self.participantId)
            except ValueError:
                ids.add(self.participantId)

        for entry in self.entries:
            if entry.playerOrTeamId:
                try:
                    int(entry.playerOrTeamId)
                except ValueError:
                    ids.add(entry.playerOrTeamId)
        return ids


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_mini_series():
    global MiniSeries

    @cassiopeia.type.core.common.inheritdocs
    class MiniSeries(MiniSeries, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MiniSeries"
        losses = sqlalchemy.Column(sqlalchemy.Integer)
        progress = sqlalchemy.Column(sqlalchemy.String(30))
        target = sqlalchemy.Column(sqlalchemy.Integer)
        wins = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _entry_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("LeagueEntry._id", ondelete="CASCADE"))


def _sa_bind_league_entry():
    global LeagueEntry

    @cassiopeia.type.core.common.inheritdocs
    class LeagueEntry(LeagueEntry, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "LeagueEntry"
        division = sqlalchemy.Column(sqlalchemy.String(30))
        isFreshBlood = sqlalchemy.Column(sqlalchemy.Boolean)
        isHotStreak = sqlalchemy.Column(sqlalchemy.Boolean)
        isInactive = sqlalchemy.Column(sqlalchemy.Boolean)
        isVeteran = sqlalchemy.Column(sqlalchemy.Boolean)
        leaguePoints = sqlalchemy.Column(sqlalchemy.Integer)
        losses = sqlalchemy.Column(sqlalchemy.Integer)
        miniSeries = sqlalchemy.orm.relationship("cassiopeia.type.dto.league.MiniSeries", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        playerOrTeamId = sqlalchemy.Column(sqlalchemy.String(50))
        playerOrTeamName = sqlalchemy.Column(sqlalchemy.String(30))
        wins = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _league_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("League._id", ondelete="CASCADE"))


def _sa_bind_league():
    global League

    @cassiopeia.type.core.common.inheritdocs
    class League(League, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "League"
        entries = sqlalchemy.orm.relationship("cassiopeia.type.dto.league.LeagueEntry", cascade="all, delete-orphan", passive_deletes=True)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        participantId = sqlalchemy.Column(sqlalchemy.String(50))
        queue = sqlalchemy.Column(sqlalchemy.String(30))
        tier = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)


def _sa_bind_all():
    _sa_bind_mini_series()
    _sa_bind_league_entry()
    _sa_bind_league()



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/currentgameapi tests...")
    test_current_game()


def test_current_game():
    int_test_handler.test_result(baseriotapi.get_current_game(int_test_handler.summoner_id))



import cassiopeia.riotapi
import cassiopeia.dto.statsapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.stats


def get_ranked_stats(summoner, season=None):
    """
    Gets a summoner's ranked stats

    Args:
        summoner (Summoner): the summoner to get ranked stats for
        season (Season): the season to get ranked stats for (None will give current season stats) (default None)

    Returns:
        dict<Champion, AggregatedStats>: the summoner's ranked stats divided by champion. The entry for None contains combined stats for all champions.
    """
    stats = cassiopeia.dto.statsapi.get_ranked_stats(summoner.id, season.value if season else None)
    champions = {champion.id: champion for champion in cassiopeia.riotapi.get_champions_by_id(list(stats.champion_ids))}
    champions[0] = None

    return {champions[stat.id]: cassiopeia.type.core.stats.AggregatedStats(stat.stats) for stat in stats.champions}


def get_stats(summoner, season=None):
    """
    Gets a summoner's stats

    Args:
        summoner (Summoner): the summoner to get stats for
        season (Season): the season to get stats for (None will give current season stats) (default None)

    Returns:
        dict<StatSummaryType, StatsSummary>: the summoner's stats divided by queue type
    """
    stats = cassiopeia.dto.statsapi.get_stats(summoner.id, season.value if season else None)
    return {cassiopeia.type.core.common.StatSummaryType(summary.playerStatSummaryType): cassiopeia.type.core.stats.StatsSummary(summary) for summary in stats.playerStatSummaries}



import cassiopeia.dto.requests
import cassiopeia.type.core.common
import cassiopeia.type.dto.championmastery


def get_champion_mastery(summoner_id, champion_id):
    """
    https://developer.riotgames.com/api/methods#!/1034/3545

    Args:
        summoner_id (int): the summoner ID to get champion mastery for
        champion_id (int): the champion ID for the desired champion

    Returns:
        list<ChampionMastery>: the summoner's champion mastery value for the specified champion
    """
    region = cassiopeia.type.core.common.Region(cassiopeia.dto.requests.region)
    platform = cassiopeia.type.core.common.Platform[region.name]

    # Get JSON response
    request = "https://{server}.api.pvp.net/championmastery/location/{platform}/player/{summonerId}/champion/{championId}".format(server=cassiopeia.dto.requests.region, platform=platform.value, summonerId=summoner_id, championId=champion_id)
    return cassiopeia.type.dto.championmastery.ChampionMastery(cassiopeia.dto.requests.get(request, include_base=False))


def get_champion_masteries(summoner_id):
    """
    https://developer.riotgames.com/api/methods#!/1034/3544

    Args:
        summoner_id (int): the summoner ID to get champion masteries for

    Returns:
        list<ChampionMastery>: the summoner's champion masteries
    """
    region = cassiopeia.type.core.common.Region(cassiopeia.dto.requests.region)
    platform = cassiopeia.type.core.common.Platform[region.name]

    # Get JSON response
    request = "https://{server}.api.pvp.net/championmastery/location/{platform}/player/{summonerId}/champions".format(server=cassiopeia.dto.requests.region, platform=platform.value, summonerId=summoner_id)
    response = cassiopeia.dto.requests.get(request, include_base=False)

    # Convert response to Dto type
    return [cassiopeia.type.dto.championmastery.ChampionMastery(cm) for cm in response]


def get_champion_mastery_score(summoner_id):
    """
    https://developer.riotgames.com/api/methods#!/1034/3546

    Args:
        summoner_id (int): the summoner ID to get champion masteries for

    Returns:
        int: the summoner's total champion mastery score
    """
    region = cassiopeia.type.core.common.Region(cassiopeia.dto.requests.region)
    platform = cassiopeia.type.core.common.Platform[region.name]

    # Get JSON response
    request = "https://{server}.api.pvp.net/championmastery/location/{platform}/player/{summonerId}/score".format(server=cassiopeia.dto.requests.region, platform=platform.value, summonerId=summoner_id)
    return cassiopeia.dto.requests.get(request, include_base=False)


def get_top_champion_masteries(summoner_id, count=3):
    """
    https://developer.riotgames.com/api/methods#!/1034/3540

    Args:
        summoner_id (int): the summoner ID to get champion masteries for
        count (int): the maximum number of entires to retrieve (default 3)

    Returns:
        list<ChampionMastery>: the summoner's top champion masteries
    """
    region = cassiopeia.type.core.common.Region(cassiopeia.dto.requests.region)
    platform = cassiopeia.type.core.common.Platform[region.name]

    # Get JSON response
    request = "https://{server}.api.pvp.net/championmastery/location/{platform}/player/{summonerId}/topchampions".format(server=cassiopeia.dto.requests.region, platform=platform.value, summonerId=summoner_id)
    response = cassiopeia.dto.requests.get(request, {"count": count}, include_base=False)

    # Convert response to Dto type
    return [cassiopeia.type.dto.championmastery.ChampionMastery(cm) for cm in response]



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class RunePages(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        pages (list<RunePage>): collection of rune pages associated with the summoner
        summonerId (int): summoner ID
    """
    def __init__(self, dictionary):
        self.pages = [(RunePage(p) if not isinstance(p, RunePage) else p) for p in dictionary.get("pages", []) if p]
        self.summonerId = dictionary.get("summonerId", 0)

    @property
    def rune_ids(self):
        """
        Gets all rune IDs contained in this object
        """
        ids = set()
        for p in self.pages:
            ids = ids | p.rune_ids
        return ids


@cassiopeia.type.core.common.inheritdocs
class RunePage(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all rune IDs contained in this object
    """
    def __init__(self, dictionary):
        self.current = dictionary.get("current", False)
        self.id = dictionary.get("id", 0)
        self.name = dictionary.get("name", "")
        self.slots = [(RuneSlot(s) if not isinstance(s, RuneSlot) else s) for s in dictionary.get("slots", []) if s]

    @property
    def rune_ids(self):
        """
        Args:
            current (bool): indicates if the page is the current page
            id (int): rune page ID
            name (str): rune page name
            slots (list<RuneSlot>): collection of rune slots associated with the rune page
        """
        ids = set()
        for s in self.slots:
            if s.runeId:
                ids.add(s.runeId)
        return ids


@cassiopeia.type.core.common.inheritdocs
class RuneSlot(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        current (bool): indicates if the page is the current page
        id (int): rune page ID
        name (str): rune page name
        slots (list<RuneSlot>): collection of rune slots associated with the rune page
    """
    def __init__(self, dictionary):
        self.runeId = dictionary.get("runeId", 0)
        self.runeSlotId = dictionary.get("runeSlotId", 0)


@cassiopeia.type.core.common.inheritdocs
class MasteryPages(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all rune IDs contained in this object
    """
    def __init__(self, dictionary):
        self.pages = [(MasteryPage(p) if not isinstance(p, MasteryPage) else p) for p in dictionary.get("pages", []) if p]
        self.summonerId = dictionary.get("summonerId", 0)

    @property
    def mastery_ids(self):
        """
        Args:
            runeId (int): rune ID associated with the rune slot. For static information correlating to rune IDs, please refer to the LoL Static Data API.
            runeSlotId (int): rune slot ID.
        """
        ids = set()
        for p in self.pages:
            ids = ids | p.mastery_ids
        return ids


@cassiopeia.type.core.common.inheritdocs
class MasteryPage(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        runeId (int): rune ID associated with the rune slot. For static information correlating to rune IDs, please refer to the LoL Static Data API.
        runeSlotId (int): rune slot ID.
    """
    def __init__(self, dictionary):
        self.current = dictionary.get("current", False)
        self.id = dictionary.get("id", 0)
        self.masteries = [(Mastery(s) if not isinstance(s, Mastery) else s) for s in dictionary.get("masteries", []) if s]
        self.name = dictionary.get("name", "")

    @property
    def mastery_ids(self):
        """
        Args:
            pages (list<MasteryPage>): collection of mastery pages associated with the summoner
            summonerId (int): summoner ID
        """
        ids = set()
        for m in self.masteries:
            if m.id:
                ids.add(m.id)
        return ids


@cassiopeia.type.core.common.inheritdocs
class Mastery(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        pages (list<MasteryPage>): collection of mastery pages associated with the summoner
        summonerId (int): summoner ID
    """
    def __init__(self, dictionary):
        self.id = dictionary.get("id", 0)
        self.rank = dictionary.get("rank", 0)


@cassiopeia.type.core.common.inheritdocs
class Summoner(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all mastery IDs contained in this object
    """
    def __init__(self, dictionary):
        self.id = dictionary.get("id", 0)
        self.name = dictionary.get("name", "")
        self.profileIconId = dictionary.get("profileIconId", 0)
        self.revisionDate = dictionary.get("revisionDate", 0)
        self.summonerLevel = dictionary.get("summonerLevel", 0)


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_rune_page():
    global RunePage

    @cassiopeia.type.core.common.inheritdocs
    class RunePage(RunePage, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "RunePage"
        current = sqlalchemy.Column(sqlalchemy.Boolean)
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        name = sqlalchemy.Column(sqlalchemy.String(50))
        slots = sqlalchemy.orm.relationship("cassiopeia.type.dto.summoner.RuneSlot", cascade="all, delete-orphan, delete, merge", passive_deletes=True)


def _sa_bind_rune_slot():
    global RuneSlot

    @cassiopeia.type.core.common.inheritdocs
    class RuneSlot(RuneSlot, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "RuneSlot"
        runeId = sqlalchemy.Column(sqlalchemy.Integer)
        runeSlotId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _page_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("RunePage.id", ondelete="CASCADE"))


def _sa_bind_mastery_page():
    global MasteryPage

    @cassiopeia.type.core.common.inheritdocs
    class MasteryPage(MasteryPage, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MasteryPage"
        current = sqlalchemy.Column(sqlalchemy.Boolean)
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        masteries = sqlalchemy.orm.relationship("cassiopeia.type.dto.summoner.Mastery", cascade="all, delete-orphan, delete, merge", passive_deletes=True)
        name = sqlalchemy.Column(sqlalchemy.String(50))


def _sa_bind_mastery():
    global Mastery

    @cassiopeia.type.core.common.inheritdocs
    class Mastery(Mastery, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MasterySlot"
        id = sqlalchemy.Column(sqlalchemy.Integer)
        rank = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _page_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MasteryPage.id", ondelete="CASCADE"))


def _sa_bind_summoner():
    global Summoner

    @cassiopeia.type.core.common.inheritdocs
    class Summoner(Summoner, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "Summoner"
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        name = sqlalchemy.Column(sqlalchemy.String(30))
        profileIconId = sqlalchemy.Column(sqlalchemy.Integer)
        revisionDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        summonerLevel = sqlalchemy.Column(sqlalchemy.Integer)


def _sa_bind_all():
    _sa_bind_rune_page()
    _sa_bind_rune_slot()
    _sa_bind_mastery_page()
    _sa_bind_mastery()
    _sa_bind_summoner()



import datetime

import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.matchlist


@cassiopeia.type.core.common.inheritdocs
class MatchReference(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.matchlist.MatchReference

    def __str__(self):
        return "Match {id} Reference".format(id=self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @cassiopeia.type.core.common.immutablemethod
    def match(self, include_timeline=True):
        """
        Gets the full information for this match

        Args:
            include_timeline (bool): whether to include timeline data in the returned match

        Returns:
            Match: the match
        """
        return cassiopeia.riotapi.get_match(self, include_timeline)

    @cassiopeia.type.core.common.lazyproperty
    def champion(self):
        """
        Returns:
            Champion: the champion that the summoner played for the summoner that was used to pull this match reference
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.champion)

    @property
    def lane(self):
        """
        Returns:
            Lane: the lane that the summoner was in for the summoner that was used to  pull this match reference
        """
        lane = self.data.lane
        lane = "MIDDLE" if lane == "MID" else lane
        lane = "BOTTOM" if lane == "BOT" else lane
        return cassiopeia.type.core.common.Lane(lane) if lane else None

    @property
    def id(self):
        """
        Returns:
            int: the match id for this match
        """
        return self.data.matchId

    @property
    def platform(self):
        """
        Returns:
            Platform: the platform (ie server) for this match
        """
        return cassiopeia.type.core.common.Platform(self.data.platformId) if self.data.platformId else None

    @property
    def queue(self):
        """
        Returns:
            Queue: the queue type for this match
        """
        return cassiopeia.type.core.common.Queue(self.data.queue) if self.data.queue else None

    @property
    def role(self):
        """
        Returns:
            Role: the role that the summoner was in for the summoner that was used to  pull this match reference
        """
        return cassiopeia.type.core.common.Role(self.data.role) if self.data.role else None

    @property
    def season(self):
        """
        Returns:
            Season: the season that this match was played in
        """
        return cassiopeia.type.core.common.Season(self.data.season) if self.data.season else None

    @cassiopeia.type.core.common.lazyproperty
    def timestamp(self):
        """
        Returns:
            datetime: the timestamp for this match
        """
        return datetime.datetime.utcfromtimestamp(self.data.timestamp / 1000) if self.data.timestamp else None


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    MatchReference.dto_type = cassiopeia.type.dto.matchlist.MatchReference



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/status tests...")
    test_shards()
    test_shard()


def test_shards():
    int_test_handler.test_result(baseriotapi.get_shards())


def test_shard():
    int_test_handler.test_result(baseriotapi.get_shard())



from unittest import TestCase

from cassiopeia import riotapi
from cassiopeia.type.api.exception import APIError


class Python2And3Tests(TestCase):

    def test_we_can_use_unicode(self):
        # We shouldn't check if str instead check if NOT Region
        riotapi.set_region(u"NA")
        assert True

    def test_super_imported_from_future(self):
        # Make sure we have from future.builtins.misc import super
        APIError("Some message", 123)
        assert True



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/championapi tests...")
    test_champion_status()
    test_champion_statuses()


def test_champion_status():
    int_test_handler.test_result(baseriotapi.get_champion_status(int_test_handler.champion_id))


def test_champion_statuses():
    int_test_handler.test_result(baseriotapi.get_champion_statuses())



import datetime

import cassiopeia.riotapi
import cassiopeia.dto.matchlistapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.matchlist


def get_match_list(summoner, num_matches=0, begin_index=0, begin_time=0, end_time=0, champions=None, ranked_queues=None, seasons=None):
    """
    Gets a summoner's match history

    Args:
        summoner (Summoner): the summoner to get match history for
        num_matches (int): the maximum number of matches to retrieve. 0 will get as many as possible. (default 0)
        begin_index (int): the game index to start from (default 0)
        begin_time (int | datetime): the begin time to use for fetching games (default 0)
        end_time (int | datetime): the end time to use for fetching games (default 0)
        champions (Champion | list<Champion>): the champion(s) to limit the results to (default None)
        ranked_queues Queue | list<Queue>: the ranked queue(s) to limit the results to (default None)
        seasons (Season | list<Season>): the season(s) to limit the results to (default None)

    Returns:
        list<MatchReference>: the summoner's match history
    """
    # Convert strings that should be enums into enums
    if isinstance(seasons, str):
        seasons = cassiopeia.type.core.common.Season(seasons.upper())
    elif isinstance(seasons, list) and all(isinstance(s, str) for s in seasons):
        seasons = [cassiopeia.type.core.common.Season(s.upper()) for s in seasons]

    if isinstance(ranked_queues, str):
        ranked_queues = cassiopeia.type.core.common.Queue(ranked_queues.upper().replace("5X5", "5x5").replace("3X3", "3x3"))
    elif isinstance(ranked_queues, list) and all(isinstance(q, str) for q in ranked_queues):
        ranked_queues = [cassiopeia.type.core.common.Queue(q.upper().replace("5X5", "5x5").replace("3X3", "3x3")) for q in ranked_queues]

    # Convert core types to API-ready types
    if isinstance(begin_time, datetime.datetime):
        epoch = datetime.datetime.utcfromtimestamp(0)
        delta = begin_time - epoch
        begin_time = int(delta.total_seconds() * 1000)
    if isinstance(end_time, datetime.datetime):
        epoch = datetime.datetime.utcfromtimestamp(0)
        delta = end_time - epoch
        end_time = int(delta.total_seconds() * 1000)

    champion_ids = [champion.id for champion in champions] if isinstance(champions, list) else champions.id if champions else None
    ranked_queues = [queue.value for queue in ranked_queues] if isinstance(ranked_queues, list) else ranked_queues.value if ranked_queues else None
    seasons = [season.value for season in seasons] if isinstance(seasons, list) else seasons.value if seasons else None

    history = cassiopeia.dto.matchlistapi.get_match_list(summoner.id, num_matches, begin_index, begin_time, end_time, champion_ids, ranked_queues, seasons)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_champions() if history.champion_ids else None

    return [cassiopeia.type.core.matchlist.MatchReference(ref) for ref in history.matches]



import cassiopeia.dto.requests
import cassiopeia.type.core.common
import cassiopeia.type.dto.currentgame
import cassiopeia.type.api.exception


def get_current_game(summoner_id):
    """
    https://developer.riotgames.com/api/methods#!/976/3336

    Args:
        summoner_id (int): the ID of the summoner to find an active game for

    Returns:
        CurrentGameInfo: the summoner's current game (or None if they aren't in one)
    """
    region = cassiopeia.type.core.common.Region(cassiopeia.dto.requests.region)
    platform = cassiopeia.type.core.common.Platform[region.name]
    request = "https://{server}.api.pvp.net/observer-mode/rest/consumer/getSpectatorGameInfo/{platform}/{summoner_id}".format(server=cassiopeia.dto.requests.region, platform=platform.value, summoner_id=summoner_id)
    try:
        return cassiopeia.type.dto.currentgame.CurrentGameInfo(cassiopeia.dto.requests.get(request, include_base=False))
    except cassiopeia.type.api.exception.APIError as e:
        if e.error_code == 404:
            return None
        raise e



try:
    from future.builtins.misc import super
except ImportError:
    pass


class CassiopeiaException(Exception):
    """Generic exception for a failure within Cassiopeia"""
    pass


class APIError(Exception):
    """
    Args:
        message (str): the error message
        error_code (int): the HTTP error code that was received
    """
    def __init__(self, message, error_code):
        super().__init__(message)
        self.message = message
        self.error_code = error_code



import cassiopeia.type.core.common
import cassiopeia.type.api.store

load_policy = cassiopeia.type.core.common.LoadPolicy.eager
data_store = cassiopeia.type.api.store.Cache()


def call_with_ensured_size(method, max_size, arg):
    """
    Breaks a list of arguments up into chunks of a maximum size and calls the given method on each chunk

    Args:
        method (function): the method to call
        max_size (int): the maximum number of arguments to include in a single call
        arg (any | list<any>): the arguments to split up

    Returns:
        list<any> | dict<any>: the combined results of the function calls on each chunk
    """
    if not isinstance(arg, list) or len(arg) <= max_size:
        return method(arg)

    results = method(arg[0:max_size])
    i = max_size

    if isinstance(results, list):
        while(i < len(arg)):
            sublist = arg[i:i + max_size]
            try:
                results = results + method(sublist)
            except:
                results = results + ([None] * len(sublist))
            i += max_size
    elif isinstance(results, dict):
        while(i < len(arg)):
            sublist = arg[i:i + max_size]
            try:
                results.update(method(sublist))
            except:
                pass
            i += max_size
    return results



import datetime

import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.currentgame


@cassiopeia.type.core.common.inheritdocs
class Participant(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.currentgame.CurrentGameParticipant

    def __str__(self):
        return "{player} ({champ})".format(player=self.summoner_name, champ=self.champion)

    @property
    def bot(self):
        """
        Returns:
            bool: whether the participant is a bot
        """
        return self.data.bot

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion this participant is playing
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.championId) if self.data.championId else None

    @cassiopeia.type.core.common.lazyproperty
    def masteries(self):
        """
        Returns:
            list<Mastery>: the participant's masteries
        """
        masteries = []
        ranks = []
        for mastery in self.data.masteries:
            masteries.append(mastery.masteryId)
            ranks.append(mastery.rank)
        return dict(zip(cassiopeia.riotapi.get_masteries(masteries), ranks))

    @property
    def profile_icon_id(self):
        """
        Returns:
            int: the participant's profile icon's id
        """
        return self.data.profileIconId

    @cassiopeia.type.core.common.lazyproperty
    def runes(self):
        """
        Returns:
            list<Rune>: the participant's rune
        """
        runes = []
        counts = []
        for rune in self.data.runes:
            runes.append(rune.runeId)
            counts.append(rune.count)
        return dict(zip(cassiopeia.riotapi.get_runes(runes), counts))

    @property
    def summoner_spell_d(self):
        """
        Returns:
            SummonerSpell: the participant's first summoner spell
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.spell1Id) if self.data.spell1Id else None

    @property
    def summoner_spell_f(self):
        """
        Returns:
            SummonerSpell: the participant's second summoner spell
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.spell2Id) if self.data.spell2Id else None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner associated with this participant
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.data.summonerId) if self.data.summonerId else None

    @property
    def summoner_name(self):
        """
        Returns:
            str: the participant's summoner name
        """
        return self.data.summonerName

    @property
    def side(self):
        """
        Returns:
            Side: which side of the map the participant is on
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.teamId else None


@cassiopeia.type.core.common.inheritdocs
class Ban(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.currentgame.BannedChampion

    def __str__(self):
        return "Ban ({champ})".format(champ=self.champion)

    @property
    def champion(self):
        """
        Returns:
            Side: which side banned this champion
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.championId) if self.data.championId else None

    @property
    def pick_turn(self):
        """
        Returns:
            int: which pick turn this ban was on
        """
        return self.data.pickturn

    @property
    def side(self):
        """
        Returns:
            Side: the team that made this ban
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.teamId else None


@cassiopeia.type.core.common.inheritdocs
class Game(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.currentgame.CurrentGameInfo

    def __str__(self):
        return "Game #{id}".format(id=self.id)

    def __iter__(self):
        return iter(self.participants)

    def __len__(self):
        return len(self.participants)

    def __getitem__(self, index):
        return self.participants[index]

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @cassiopeia.type.core.common.lazyproperty
    def bans(self):
        """
        Returns:
            list<Ban>: the bans for this game
        """
        return [Ban(ban) for ban in self.data.bannedChampions]

    @property
    def id(self):
        """
        Returns:
            int: the game id
        """
        return self.data.gameId

    @cassiopeia.type.core.common.lazyproperty
    def duration(self):
        """
        Returns:
            timedelta: current duration of the game
        """
        return datetime.timedelta(seconds=self.data.gameLength)

    @property
    def mode(self):
        """
        Returns:
            GameMode: what game mode is being played in this game
        """
        return cassiopeia.type.core.common.GameMode(self.data.gameMode) if self.data.gameMode else None

    @property
    def queue(self):
        """
        Returns:
            Queue: the queue type for this game
        """
        return cassiopeia.type.core.common.Queue.for_id(self.data.gameQueueConfigId) if self.data.gameQueueConfigId else None

    @cassiopeia.type.core.common.lazyproperty
    def creation(self):
        """
        Returns:
            datetime: the creation timestamp for this game
        """
        return datetime.datetime.utcfromtimestamp(self.data.gameStartTime / 1000) if self.data.gameStartTime else None

    @property
    def type(self):
        """
        Returns:
            GameType: the game type
        """
        return cassiopeia.type.core.common.GameType(self.data.gameType) if self.data.gameType else None

    @property
    def map(self):
        """
        Returns:
            Map: the map for this game
        """
        return cassiopeia.type.core.common.Map(self.data.mapId) if self.data.mapId else None

    @property
    def observer_token(self):
        """
        Returns:
            str: the token associated with the observer for this game
        """
        return self.data.observers.encryptionKey

    @cassiopeia.type.core.common.lazyproperty
    def participants(self):
        """
        Returns:
            list<Participant>: the game's participants
        """
        return [Participant(participant) for participant in self.data.participants]

    @property
    def platform(self):
        """
        Returns:
            Platform: which platform (ie server) the game is being played on
        """
        return cassiopeia.type.core.common.Platform(self.data.platformId) if self.data.platformId else None


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Participant.dto_type = cassiopeia.type.dto.currentgame.CurrentGameParticipant
    Ban.dto_type = cassiopeia.type.dto.currentgame.BannedChampion
    Game.dto_type = cassiopeia.type.dto.currentgame.CurrentGameInfo



import cassiopeia.dto.requests
import cassiopeia.type.dto.match


def get_match(id_, include_timeline=True, tournament_code=""):
    """
    https://developer.riotgames.com/api/methods#!/1014/3442

    Args:
        id_ (int): the ID of the match to get
        include_timeline (bool): whether to include timeline data in the returned match
        tournament_code (str): the tournament code if the match to be retrieved is from a tournament

    Returns:
        MatchDetail: the match
    """
    request = "{version}/match/for-tournament/{id_}" if tournament_code else "{version}/match/{id_}"
    request = request.format(version=cassiopeia.dto.requests.api_versions["match"], id_=id_)

    params = {"includeTimeline": "true" if include_timeline else "false"}
    if tournament_code:
        params["tournamentCode"] = tournament_code

    return cassiopeia.type.dto.match.MatchDetail(cassiopeia.dto.requests.get(request, params, tournament=bool(tournament_code)))


def get_tournament_match_ids(tournament_code):
    """
    https://developer.riotgames.com/api/methods#!/1058/3656

    Args:
        tournament_code (str): the tournament code

    Returns:
        list<int>: the match ids for the tournament
    """
    request = "{version}/match/by-tournament/{tournament_code}/ids".format(version=cassiopeia.dto.requests.api_versions["match"], tournament_code=tournament_code)
    return cassiopeia.dto.requests.get(request, tournament=True)



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/status tests...")
    test_shards()
    test_shard()


def test_shards():
    int_test_handler.test_result(riotapi.get_shards())


def test_shard():
    int_test_handler.test_result(riotapi.get_shard())



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy


@cassiopeia.type.core.common.inheritdocs
class MatchList(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        endIndex (int): the last match index from history returned
        matches (list<MatchReference>): list of matches from the player's history
        startIndex (int): the first match index from history returned
        totalGames (int): the number of games provided
    """
    def __init__(self, dictionary):
        self.endIndex = dictionary.get("endIndex", 0)
        self.matches = [(MatchReference(match) if not isinstance(match, MatchReference) else match) for match in dictionary.get("matches", []) if match]
        self.startIndex = dictionary.get("startIndex", 0)
        self.totalGames = dictionary.get("totalGames", 0)

    @property
    def champion_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for m in self.matches:
            ids.add(m.champion)
        return ids


@cassiopeia.type.core.common.inheritdocs
class MatchReference(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all champion IDs contained in this object
    """
    def __init__(self, dictionary):
        self.champion = dictionary.get("champion", 0)
        self.lane = dictionary.get("lane", "")
        self.matchId = dictionary.get("matchId", 0)
        self.platformId = dictionary.get("platformId", "")
        self.queue = dictionary.get("queue", "")
        self.role = dictionary.get("role", "")
        self.season = dictionary.get("season", "")
        self.timestamp = dictionary.get("timestamp", 0)


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_match_reference():
    global MatchReference

    @cassiopeia.type.core.common.inheritdocs
    class MatchReference(MatchReference, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchReference"
        champion = sqlalchemy.Column(sqlalchemy.Integer)
        lane = sqlalchemy.Column(sqlalchemy.String(30))
        matchId = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        platformId = sqlalchemy.Column(sqlalchemy.String(30))
        queue = sqlalchemy.Column(sqlalchemy.String(30))
        role = sqlalchemy.Column(sqlalchemy.String(30))
        season = sqlalchemy.Column(sqlalchemy.String(30))
        timestamp = sqlalchemy.Column(sqlalchemy.BigInteger)


def _sa_bind_all():
    _sa_bind_match_reference()



import datetime

import cassiopeia.riotapi
import cassiopeia.type.dto.common
import cassiopeia.type.core.common
import cassiopeia.type.dto.match

try:
    from future.builtins.misc import super
except ImportError:
    pass


@cassiopeia.type.core.common.inheritdocs
class Match(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.MatchDetail

    def __str__(self):
        return "Match #{id}".format(id=self.id)

    def __iter__(self):
        return iter(self.participants)

    def __len__(self):
        return len(self.participants)

    def __getitem__(self, index):
        return self.participants[index]

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def map(self):
        """
        Returns:
            Map: the map the match was played on
        """
        return cassiopeia.type.core.common.Map(self.data.mapId) if self.data.mapId else None

    @cassiopeia.type.core.common.lazyproperty
    def creation(self):
        """
        Returns:
            datetime: when the match was created
        """
        return datetime.datetime.utcfromtimestamp(self.data.matchCreation / 1000) if self.data.matchCreation else None

    @cassiopeia.type.core.common.lazyproperty
    def duration(self):
        """
        Returns:
            datetime: duration of the match
        """
        return datetime.timedelta(seconds=self.data.matchDuration)

    @property
    def id(self):
        """
        Returns:
            int: the match ID
        """
        return self.data.matchId

    @property
    def mode(self):
        """
        Returns:
            GameMode: the game mode
        """
        return cassiopeia.type.core.common.GameMode(self.data.matchMode) if self.data.matchMode else None

    @property
    def type(self):
        """
        Returns:
            GameType: the game type
        """
        return cassiopeia.type.core.common.GameType(self.data.matchType) if self.data.matchType else None

    @property
    def version(self):
        """
        Returns:
            str: the patch this match was played in
        """
        return self.data.matchVersion

    @cassiopeia.type.core.common.lazyproperty
    def participants(self):
        """
        Returns:
            list<Participant>: the participants in this match
        """
        participants = []
        for i in range(len(self.data.participants)):
            p = CombinedParticipant(self.data.participants[i], self.data.participantIdentities[i])
            participants.append(Participant(p))
        return sorted(participants, key=lambda p: p.id)

    @property
    def platform(self):
        """
        Returns:
            Platform: the platform (ie server) for this match
        """
        return cassiopeia.type.core.common.Platform(self.data.platformId) if self.data.platformId else None

    @property
    def queue(self):
        """
        Returns:
            Queue: the queue type for this match
        """
        return cassiopeia.type.core.common.Queue(self.data.queueType) if self.data.queueType else None

    @property
    def region(self):
        """
        Returns:
            Region: the region the match was played in
        """
        return cassiopeia.type.core.common.Region(self.data.region.lower()) if self.data.region else None

    @property
    def season(self):
        """
        Returns:
            Season: the season this match was played in
        """
        return cassiopeia.type.core.common.Season(self.data.season) if self.data.season else None

    @cassiopeia.type.core.common.lazyproperty
    def blue_team(self):
        """
        Returns:
            Team: the blue team
        """
        for team in self.data.teams:
            if team.teamId == cassiopeia.type.core.common.Side.blue.value:
                return Team(team, [part for part in self.participants if part.side is cassiopeia.type.core.common.Side.blue])
        return None

    @cassiopeia.type.core.common.lazyproperty
    def red_team(self):
        """
        Returns:
            Team: the red team
        """
        for team in self.data.teams:
            if team.teamId == cassiopeia.type.core.common.Side.red.value:
                return Team(team, [part for part in self.participants if part.side is cassiopeia.type.core.common.Side.red])
        return None

    @cassiopeia.type.core.common.lazyproperty
    def timeline(self):
        """
        Returns:
            Timeline: the match timeline
        """
        return Timeline(self.data.timeline, self.participants) if self.data.timeline else None

    @cassiopeia.type.core.common.lazyproperty
    def frames(self):
        """
        Returns:
            list<Frame>: the frames in this match
        """
        return self.timeline.frames


@cassiopeia.type.core.common.inheritdocs
class CombinedParticipant(cassiopeia.type.dto.common.CassiopeiaDto):
    def __init__(self, participant, identity):
        self.participant = participant
        self.identity = identity


@cassiopeia.type.core.common.inheritdocs
class Participant(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = CombinedParticipant

    def __str__(self):
        return "{player} ({champ})".format(player=self.summoner_name, champ=self.champion)

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion this participant played
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.participant.championId) if self.data.participant.championId else None

    @property
    def previous_season_tier(self):
        """
        Returns:
            Tier: the participant's tier last season
        """
        return cassiopeia.type.core.common.Tier(self.data.participant.highestAchievedSeasonTier) if self.data.participant.highestAchievedSeasonTier else None

    @cassiopeia.type.core.common.lazyproperty
    def masteries(self):
        """
        Returns:
            list<Mastery>: the participant's masteries
        """
        masteries = []
        ranks = []
        for mastery in self.data.participant.masteries:
            masteries.append(mastery.masteryId)
            ranks.append(mastery.rank)
        return dict(zip(cassiopeia.riotapi.get_masteries(masteries), ranks))

    @property
    def id(self):
        """
        Returns:
            int: the participant ID
        """
        return self.data.participant.participantId

    @cassiopeia.type.core.common.lazyproperty
    def runes(self):
        """
        Returns:
            list<Rune>: the participant's current runes
        """
        runes = []
        counts = []
        for rune in self.data.participant.runes:
            runes.append(rune.runeId)
            counts.append(rune.rank)
        return dict(zip(cassiopeia.riotapi.get_runes(runes), counts))

    @property
    def summoner_spell_d(self):
        """
        Returns:
            SummonerSpell: the participant's first summoner spell
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.participant.spell1Id) if self.data.participant.spell1Id else None

    @property
    def summoner_spell_f(self):
        """
        Returns:
            SummonerSpell: the participant's second summoner spell
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.participant.spell2Id) if self.data.participant.spell2Id else None

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            ParticipantStats: the participant's stats
        """
        return ParticipantStats(self.data.participant.stats) if self.data.participant.stats else None

    @property
    def side(self):
        """
        Returns:
            Side: the side this participant was on
        """
        return cassiopeia.type.core.common.Side(self.data.participant.teamId) if self.data.participant.teamId else None

    @cassiopeia.type.core.common.lazyproperty
    def timeline(self):
        """
        Returns:
            ParticipantTimeline: the participant's timeline
        """
        return ParticipantTimeline(self.data.participant.timeline) if self.data.participant.timeline else None

    @property
    def match_history_uri(self):
        """
        Returns:
            str: the the URI to access this player's match history online
        """
        return self.data.identity.player.matchHistoryUri if self.data.identity.player else None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner associated with this participant
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.data.identity.player.summonerId) if self.data.identity.player and self.data.identity.player.summonerId else None

    @property
    def summoner_id(self):
        """
        Returns:
            str: the participant's summoner id
        """
        return self.data.identity.player.summonerId if self.data.identity.player else None

    @property
    def summoner_name(self):
        """
        Returns:
            str: the participant's summoner name
        """
        return self.data.identity.player.summonerName if self.data.identity.player else None


@cassiopeia.type.core.common.inheritdocs
class Team(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.Team

    def __init__(self, data, participants):
        super().__init__(data)
        self.__participants = participants

    def __str__(self):
        return "{side} team: {players}".format(side=self.side, players=self.__participants)

    def __iter__(self):
        return iter(self.__participants)

    def __len__(self):
        return len(self.__participants)

    def __getitem__(self, index):
        return self.__participants[index]

    @property
    def participants(self):
        """
        Returns:
            list<Participant>: the participants on this team
        """
        return self.__participants

    @cassiopeia.type.core.common.lazyproperty
    def bans(self):
        """
        Returns:
            list<Ban>: the bans for this game
        """
        return [Ban(ban) for ban in self.data.bans]

    @property
    def baron_kills(self):
        """
        Returns:
            int: the number of times the team killed Baron
        """
        return self.data.baronKills

    @property
    def victory_score(self):
        """
        Returns:
            int: dominion only, the points the team had at game end
        """
        return self.data.dominionVictoryScore

    @property
    def dragon_kills(self):
        """
        Returns:
            int: the number of times the team killed Dragon
        """
        return self.data.dragonKills

    @property
    def first_baron(self):
        """
        Returns:
            bool: whether or not the team killed the first baron
        """
        return self.data.firstBaron

    @property
    def first_blood(self):
        """
        Returns:
            bool: whether this team got first blood
        """
        return self.data.firstBlood

    @property
    def first_dragon(self):
        """
        Returns:
            bool: whether or not this team killed the first dragon
        """
        return self.data.firstDragon

    @property
    def first_inhibitor(self):
        """
        Returns:
            bool: flag indicating if this team destroyed the first inhibitor
        """
        return self.data.firstInhibitor

    @property
    def first_rift_herald(self):
        """
        Returns:
            bool: flag indicating if this team killed the first rift herald
        """
        return self.data.firstRiftHerald

    @property
    def first_turret(self):
        """
        Returns:
            bool: flag indicating if this team destroyed the first tower
        """
        return self.data.firstTower

    @property
    def inhibitor_kills(self):
        """
        Returns:
            int: the number of inhibitors this team killed
        """
        return self.data.inhibitorKills

    @property
    def rift_herald_kills(self):
        """
        Returns:
            int: the number of rift heralds this team killed
        """
        return self.data.riftHeraldKills

    @property
    def side(self):
        """
        Returns:
            Side: the side this team was on
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.teamId else None

    @property
    def turret_kills(self):
        """
        Returns:
            int: the number of turret kills this team had
        """
        return self.data.towerKills

    @property
    def vilemaw_kills(self):
        """
        Returns:
            int: the number of times the team has killed Vilemaw
        """
        return self.data.vilemawKills

    @property
    def win(self):
        """
        Returns:
            bool: whether or not the team won the game
        """
        return self.data.winner


@cassiopeia.type.core.common.inheritdocs
class Timeline(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.Timeline

    def __init__(self, data, participants):
        super().__init__(data)
        self.__participants = participants

    def __str__(self):
        return "Timeline"

    def __iter__(self):
        return iter(self.frames)

    def __len__(self):
        return len(self.frames)

    def __getitem__(self, index):
        return self.frames[index]

    @cassiopeia.type.core.common.lazyproperty
    def frame_interval(self):
        """
        Returns:
            timedelta: the number of milliseconds between frames
        """
        return datetime.timedelta(milliseconds=self.data.frameInterval)

    @cassiopeia.type.core.common.lazyproperty
    def frames(self):
        """
        Returns:
            list<Frame>: the frames in this match
        """
        participants = {participant.id: participant for participant in self.__participants}
        value = [Frame(frame, participants) for frame in self.data.frames]
        del self.__participants
        return value


@cassiopeia.type.core.common.inheritdocs
class ParticipantStats(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.ParticipantStats

    def __str__(self):
        return "Participant Stats"

    @property
    def kda(self):
        """
        Returns:
            float: the participant's kda
        """
        return (self.kills + self.assists) / (self.deaths if self.deaths else 1)

    @property
    def assists(self):
        """
        Returns:
            int: the total number of assists this participant had
        """
        return self.data.assists

    @property
    def champion_level(self):
        """
        Returns:
            int: the champion level of the participant when the game ended
        """
        return self.data.champLevel

    @property
    def combat_score(self):
        """
        Returns:
            int: dominion only. the part of the participant's score that came from combat-related activities
        """
        return self.data.combatPlayerScore

    @property
    def cs(self):
        return self.minion_kills + self.monster_kills

    @property
    def deaths(self):
        """
        Returns:
            int: the number of deaths this participant had
        """
        return self.data.deaths

    @property
    def double_kills(self):
        """
        Returns:
            int: the number of double kills this participant had
        """
        return self.data.doubleKills

    @property
    def first_blood_assist(self):
        """
        Returns:
            bool: flag indicating if participant got an assist on first blood
        """
        return self.data.firstBloodAssist

    @property
    def first_blood(self):
        """
        Returns:
            bool: whether participant team got first blood
        """
        return self.data.firstBloodKill

    @property
    def first_inhibitor_assist(self):
        """
        Returns:
            bool: flag indicating if participant got an assist on the first inhibitor
        """
        return self.data.firstInhibitorAssist

    @property
    def first_inhibitor(self):
        """
        Returns:
            bool: flag indicating if this participant destroyed the first inhibitor
        """
        return self.data.firstInhibitorKill

    @property
    def first_turret_assist(self):
        """
        Returns:
            bool: flag indicating if participant got an assist on the first tower
        """
        return self.data.firstTowerAssist

    @property
    def first_turret(self):
        """
        Returns:
            bool: flag indicating if this team destroyed the first tower
        """
        return self.data.firstTowerKill

    @property
    def gold_earned(self):
        """
        Returns:
            int: the participant's total gold
        """
        return self.data.goldEarned

    @property
    def gold_spent(self):
        """
        Returns:
            int: the participant's spent gold
        """
        return self.data.goldSpent

    @property
    def inhibitor_kills(self):
        """
        Returns:
            int: the number of inhibitors this team killed
        """
        return self.data.inhibitorKills

    @property
    def item0(self):
        """
        Returns:
            Item: the participant's first item
        """
        return cassiopeia.riotapi.get_item(self.data.item0) if self.data.item0 else None

    @property
    def item1(self):
        """
        Returns:
            Item: the participant's second item
        """
        return cassiopeia.riotapi.get_item(self.data.item1) if self.data.item1 else None

    @property
    def item2(self):
        """
        Returns:
            Item: the participant's third item
        """
        return cassiopeia.riotapi.get_item(self.data.item2) if self.data.item2 else None

    @property
    def item3(self):
        """
        Returns:
            Item: the participant's fourth item
        """
        return cassiopeia.riotapi.get_item(self.data.item3) if self.data.item3 else None

    @property
    def item4(self):
        """
        Returns:
            Item: the participant's fifth item
        """
        return cassiopeia.riotapi.get_item(self.data.item4) if self.data.item4 else None

    @property
    def item5(self):
        return cassiopeia.riotapi.get_item(self.data.item5) if self.data.item5 else None

    @property
    def item6(self):
        """
        Returns:
            Item: the participant's seventh item (i.e. their ward)
        """
        return cassiopeia.riotapi.get_item(self.data.item6) if self.data.item6 else None

    @property
    def items(self):
        """
        Returns:
            list<Item>: the participant's items
        """
        return [self.item0, self.item1, self.item2, self.item3, self.item4, self.item5, self.item6]

    @property
    def killing_sprees(self):
        """
        Returns:
            int: the number of killing sprees this participant had
        """
        return self.data.killingSprees

    @property
    def kills(self):
        """
        Returns:
            int: the total number of kills this participant had
        """
        return self.data.kills

    @property
    def largest_critical_strike(self):
        """
        Returns:
            int: the largest critical strike this participant had
        """
        return self.data.largestCriticalStrike

    @property
    def largest_killing_spree(self):
        """
        Returns:
            int: the larges killing spree this participant had
        """
        return self.data.largestKillingSpree

    @property
    def largest_multi_kill(self):
        """
        Returns:
            int: the largest multikill this participant had
        """
        return self.data.largestMultiKill

    @property
    def magic_damage_dealt(self):
        """
        Returns:
            int: the total magic damage this participant dealt
        """
        return self.data.magicDamageDealt

    @property
    def magic_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total magic damage this participant dealt to champions
        """
        return self.data.magicDamageDealtToChampions

    @property
    def magic_damage_taken(self):
        """
        Returns:
            int: the total magic damage this participant received
        """
        return self.data.magicDamageTaken

    @property
    def minion_kills(self):
        """
        Returns:
            int: the number of minions killed
        """
        return self.data.minionsKilled

    @property
    def monster_kills(self):
        """
        Returns:
            int: the number of neutral minions this participant killed
        """
        return self.data.neutralMinionsKilled

    @property
    def enemy_monster_kills(self):
        """
        Returns:
            int: the number of neutral jungle minions killed in the enemy team's jungle
        """
        return self.data.neutralMinionsKilledEnemyJungle

    @property
    def ally_monster_kills(self):
        """
        Returns:
            int: the number of neutral jungle minions killed in your team's jungle
        """
        return self.data.neutralMinionsKilledTeamJungle

    @property
    def nodes_captured(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant captured
        """
        return self.data.nodeCapture

    @property
    def node_capture_assists(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant assisted in capturing
        """
        return self.data.nodeCaptureAssist

    @property
    def node_neutralizations(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant neutralized
        """
        return self.data.nodeNeutralize

    @property
    def node_neutralization_assists(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant assisted in neutralizing
        """
        return self.data.nodeNeutralizeAssist

    @property
    def objective_score(self):
        """
        Returns:
            int: dominion only. the part of the participant's score that came from objective-related activities
        """
        return self.data.objectivePlayerScore

    @property
    def penta_kills(self):
        """
        Returns:
            int: the number of penta kills this participant had
        """
        return self.data.pentaKills

    @property
    def physical_damage_dealt(self):
        """
        Returns:
            int: the total physical damage this participant dealt
        """
        return self.data.physicalDamageDealt

    @property
    def physical_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total physical damage this participant dealt to champions
        """
        return self.data.physicalDamageDealtToChampions

    @property
    def physical_damage_taken(self):
        """
        Returns:
            int: the total physical damage this participant received
        """
        return self.data.physicalDamageTaken

    @property
    def quadra_kills(self):
        """
        Returns:
            int: the number of quadra kills this participant had
        """
        return self.data.quadraKills

    @property
    def sight_wards_bought(self):
        """
        Returns:
            int: the number of sight wards this participant bought
        """
        return self.data.sightWardsBoughtInGame

    @property
    def team_objectives(self):
        """
        Returns:
            int: if game was a dominion game, number of completed team objectives (i.e., quests)
        """
        return self.data.teamObjective

    @property
    def damage_dealt(self):
        """
        Returns:
            int: the total damage this participant dealt
        """
        return self.data.totalDamageDealt

    @property
    def damage_dealt_to_champions(self):
        """
        Returns:
            int: the total damage this participant dealt to champions
        """
        return self.data.totalDamageDealtToChampions

    @property
    def damage_taken(self):
        """
        Returns:
            int: the total damage this participant received
        """
        return self.data.totalDamageTaken

    @property
    def healing_done(self):
        """
        Returns:
            int: the amount of healing this participant did
        """
        return self.data.totalHeal

    @property
    def score(self):
        """
        Returns:
            int: dominion only. the score for this participant
        """
        return self.data.totalPlayerScore

    @property
    def score_rank(self):
        """
        Returns:
            int: if game was a dominion game, team rank of the player's total score (e.g., 1-5)
        """
        return self.data.totalScoreRank

    @property
    def crowd_control_dealt(self):
        """
        Returns:
            int: the total amount of crowd control this participant dealt (in seconds)
        """
        return self.data.totalTimeCrowdControlDealt

    @property
    def units_healed(self):
        """
        Returns:
            int: the number of units this participant healed
        """
        return self.data.totalUnitsHealed

    @property
    def turret_kills(self):
        """
        Returns:
            int: the number of turret kills this participant had
        """
        return self.data.towerKills

    @property
    def triple_kills(self):
        """
        Returns:
            int: the number of triple kills this participant had
        """
        return self.data.tripleKills

    @property
    def true_damage_dealt(self):
        """
        Returns:
            int: the total true damage this participant dealth
        """
        return self.data.trueDamageDealt

    @property
    def true_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total damage this participant dealt to champions
        """
        return self.data.trueDamageDealtToChampions

    @property
    def true_damage_taken(self):
        """
        Returns:
            int: the total true damage this participant received
        """
        return self.data.trueDamageTaken

    @property
    def unreal_kills(self):
        """
        Returns:
            int: the number of unreal kills this participant had
        """
        return self.data.unrealKills

    @property
    def vision_wards_bought(self):
        """
        Returns:
            int: the number of vision wards sprees this participant bought
        """
        return self.data.visionWardsBoughtInGame

    @property
    def ward_kills(self):
        """
        Returns:
            int: the number of wards this participant killed
        """
        return self.data.wardsKilled

    @property
    def wards_placed(self):
        """
        Returns:
            int: the number of wards this participant placed
        """
        return self.data.wardsPlaced

    @property
    def win(self):
        """
        Returns:
            bool: whether or not the participant won the game
        """
        return self.data.winner


@cassiopeia.type.core.common.inheritdocs
class ParticipantTimeline(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.ParticipantTimeline

    def __str__(self):
        return "Participant Timeline"

    @cassiopeia.type.core.common.lazyproperty
    def ancient_golem_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: ancient golem assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.ancientGolemAssistsPerMinCounts) if self.data.ancientGolemAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def ancient_golem_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: ancient golem kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.ancientGolemKillsPerMinCounts) if self.data.ancientGolemKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def assisted_lane_deaths_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: assisted lane deaths per minute timeline data
        """
        return ParticipantTimelineData(self.data.assistedLaneDeathsPerMinDeltas) if self.data.assistedLaneDeathsPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def assisted_lane_kills_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: assisted lane kills per minute timeline data
        """
        return ParticipantTimelineData(self.data.assistedLaneKillsPerMinDeltas) if self.data.assistedLaneKillsPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def baron_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: baron assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.baronAssistsPerMinCounts) if self.data.baronAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def baron_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: baron kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.baronKillsPerMinCounts) if self.data.baronKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def creeps_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: creeps per minute timeline data
        """
        return ParticipantTimelineData(self.data.creepsPerMinDeltas) if self.data.creepsPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def cs_diff_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: creep score difference per minute timeline data
        """
        return ParticipantTimelineData(self.data.csDiffPerMinDeltas) if self.data.csDiffPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def damage_taken_diff_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: damage taken difference per minute timeline data
        """
        return ParticipantTimelineData(self.data.damageTakenDiffPerMinDeltas) if self.data.damageTakenDiffPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def damage_taken_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: damage taken per minute timeline data
        """
        return ParticipantTimelineData(self.data.damageTakenPerMinDeltas) if self.data.damageTakenPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def dragon_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: dragon assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.dragonAssistsPerMinCounts) if self.data.dragonAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def dragon_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: dragon kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.dragonKillsPerMinCounts) if self.data.dragonKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def elder_lizard_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: elder lizard assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.elderLizardAssistsPerMinCounts) if self.data.elderLizardAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def elder_lizard_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: elder lizard kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.elderLizardKillsPerMinCounts) if self.data.elderLizardKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def gold_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: gold per minute timeline data
        """
        return ParticipantTimelineData(self.data.goldPerMinDeltas) if self.data.goldPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def inhibitor_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: inhibitor assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.inhibitorAssistsPerMinCounts) if self.data.inhibitorAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def inhibitor_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: inhibitor kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.inhibitorKillsPerMinCounts) if self.data.inhibitorKillsPerMinCounts else None

    @property
    def lane(self):
        """
        Returns:
            Lane: the lane this participant was in
        """
        lane = self.data.lane
        lane = "MIDDLE" if lane == "MID" else lane
        lane = "BOTTOM" if lane == "BOT" else lane
        return cassiopeia.type.core.common.Lane(lane) if lane else None

    @property
    def role(self):
        """
        Returns:
            Role: the role of this particiant
        """
        return cassiopeia.type.core.common.Role(self.data.role) if self.data.role else None

    @cassiopeia.type.core.common.lazyproperty
    def turret_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: tower assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.towerAssistsPerMinCounts) if self.data.towerAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def turret_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: tower kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.towerKillsPerMinCounts) if self.data.towerKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def turret_Kills_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: tower kills per minute timeline data
        """
        return ParticipantTimelineData(self.data.towerKillsPerMinDeltas) if self.data.towerKillsPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def spider_assists_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: vilemaw assists per minute timeline counts
        """
        return ParticipantTimelineData(self.data.vilemawAssistsPerMinCounts) if self.data.vilemawAssistsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def spider_kills_per_min_counts(self):
        """
        Returns:
            ParticipantTimelineData: vilemaw kills per minute timeline counts
        """
        return ParticipantTimelineData(self.data.vilemawKillsPerMinCounts) if self.data.vilemawKillsPerMinCounts else None

    @cassiopeia.type.core.common.lazyproperty
    def wards_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: wards placed per minute timeline data
        """
        return ParticipantTimelineData(self.data.wardsPerMinDeltas) if self.data.wardsPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def xp_diff_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: experience difference per minute timeline data
        """
        return ParticipantTimelineData(self.data.xpDiffPerMinDeltas) if self.data.xpDiffPerMinDeltas else None

    @cassiopeia.type.core.common.lazyproperty
    def xp_per_min_deltas(self):
        """
        Returns:
            ParticipantTimelineData: experience per minute timeline data
        """
        return ParticipantTimelineData(self.data.xpPerMinDeltas) if self.data.xpPerMinDeltas else None


@cassiopeia.type.core.common.inheritdocs
class Ban(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.BannedChampion

    def __str__(self):
        return "Ban({champ})".format(champ=self.champion)

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion that was banned
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.championId) if self.data.championId else None

    @property
    def pick_turn(self):
        """
        Returns:
            int: which pick turn this ban was on
        """
        return self.data.pickTurn


@cassiopeia.type.core.common.inheritdocs
class Frame(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.Frame
    __participant_quota = 2

    def __init__(self, data, participants):
        super().__init__(data)
        self.__participants = participants
        self.__counter = 0

    def __str__(self):
        return "Frame ({time})".format(time=self.timestamp)

    def __iter__(self):
        return iter(self.events)

    def __len__(self):
        return len(self.events)

    def __getitem__(self, index):
        return self.events[index]

    def __count_participant(self):
        self.__counter += 1
        if self.__counter >= Frame.__participant_quota:
            del self.__counter
            del self.__participants

    @cassiopeia.type.core.common.lazyproperty
    def events(self):
        """
        Returns:
            list<Event>: the events in this frame
        """
        value = [Event(event, self.__participants) for event in self.data.events]
        self.__count_participant()
        return value

    @cassiopeia.type.core.common.lazyproperty
    def participant_frames(self):
        """
        Returns:
            dict<participantID, ParticipantFrame>: the frames in for each participant
        """
        value = {participant: ParticipantFrame(self.data.participantFrames[str(id_)], self.__participants) for id_, participant in self.__participants.items()}
        self.__count_participant()
        return value

    @cassiopeia.type.core.common.lazyproperty
    def timestamp(self):
        """
        Returns:
            datetime: the timestamp for this frame
        """
        return datetime.timedelta(milliseconds=self.data.timestamp)


@cassiopeia.type.core.common.inheritdocs
class ParticipantTimelineData(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.ParticipantTimelineData

    def __str__(self):
        return "Participant Timeline Data"

    @property
    def ten_to_twenty(self):
        """
        Returns:
            float: value per minute from 10 min to 20 min
        """
        return self.data.tenToTwenty

    @property
    def thirty_to_end(self):
        """
        Returns:
            float: value per minute from 30 min to the end of the game
        """
        return self.data.thirtyToEnd

    @property
    def twenty_to_thirty(self):
        """
        Returns:
            float: value per minute from 20 min to 30 min
        """
        return self.data.twentyToThirty

    @property
    def zero_to_ten(self):
        """
        Returns:
            float: value per minute from the beginning of the game to 10 min
        """
        return self.data.zeroToTen


@cassiopeia.type.core.common.inheritdocs
class Event(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.Event
    __participant_quota = 5

    def __str__(self):
        return "Event ({type})".format(type=self.type)

    def __init__(self, data, participants):
        super().__init__(data)
        self.__participants = participants
        self.__counter = 0

    def __count_participant(self):
        self.__counter += 1
        if self.__counter >= Event.__participant_quota:
            del self.__counter
            del self.__participants

    @property
    def ascended(self):
        """
        Returns:
            Ascended: what died in the event
        """
        return cassiopeia.type.core.common.Ascended(self.data.ascendedType) if self.data.ascendedType else None

    @cassiopeia.type.core.common.lazyproperty
    def assists(self):
        """
        Returns:
            list<Participant>: the participants who assisted in the event
        """
        value = [self.__participants[i] for i in self.data.assistingParticipantIds]
        self.__count_participant()
        return value

    @property
    def building(self):
        """
        Returns:
            Building: the building type associated with the event, if any
        """
        return cassiopeia.type.core.common.Building(self.data.buildingType) if self.data.buildingType else None

    @cassiopeia.type.core.common.lazyproperty
    def creator(self):
        """
        Returns:
            Participant: the participant who created the event
        """
        value = self.__participants[self.data.creatorId] if self.data.creatorId else None
        self.__count_participant()
        return value

    @property
    def type(self):
        """
        Returns:
            EventType: the event type
        """
        return cassiopeia.type.core.common.EventType(self.data.eventType) if self.data.eventType else None

    @property
    def item_after(self):
        """
        Returns:
            Item: the item involved before the event happened
        """
        return cassiopeia.riotapi.get_item(self.data.itemAfter) if self.data.itemAfter else None

    @property
    def item_before(self):
        """
        Returns:
            Item: the item involved after the event happened
        """
        return cassiopeia.riotapi.get_item(self.data.itemBefore) if self.data.itemBefore else None

    @property
    def item(self):
        """
        Returns:
            Item: the item involved in the event
        """
        return cassiopeia.riotapi.get_item(self.data.itemId) if self.data.itemId else None

    @cassiopeia.type.core.common.lazyproperty
    def killer(self):
        """
        Returns:
            Participant: the participant who did the killing
        """
        value = self.__participants[self.data.killerId] if self.data.killerId else None
        self.__count_participant()
        return value

    @property
    def lane(self):
        """
        Returns:
            Lane: the lane this event happened in
        """
        return cassiopeia.type.core.common.LaneType(self.data.laneType) if self.data.laneType else None

    @property
    def level_up(self):
        """
        Returns:
            LevelUp: the level up type of the event
        """
        return cassiopeia.type.core.common.LevelUp(self.data.levelUpType) if self.data.levelUpType else None

    @property
    def monster(self):
        """
        Returns:
            Monster: the monster that was involved in the event
        """
        return cassiopeia.type.core.common.Monster(self.data.monsterType) if self.data.monsterType else None

    @cassiopeia.type.core.common.lazyproperty
    def participant(self):
        """
        Returns:
            Participant: the primary participant that event happened to or who was involved in the event
        """
        value = self.__participants[self.data.participantId] if self.data.participantId else None
        self.__count_participant()
        return value

    @property
    def point_captured(self):
        """
        Returns:
            Point: dominion only, which point was captured
        """
        return cassiopeia.type.core.common.Point(self.data.pointCaptured) if self.data.pointCaptured else None

    @cassiopeia.type.core.common.lazyproperty
    def position(self):
        """
        Returns:
            Position: the position where the event occurred
        """
        return Position(self.data.position) if self.data.position else None

    @property
    def skill_slot(self):
        """
        Returns:
            int: the skill slot of the event
        """
        return self.data.skillSlot

    @property
    def tower(self):
        """
        Returns:
            Tower: which tower was involved in the event
        """
        return cassiopeia.type.core.common.Turret(self.data.towerType) if self.data.teamId else None

    @cassiopeia.type.core.common.lazyproperty
    def timestamp(self):
        """
        Returns:
            datetime: the timestamp for this event
        """
        return datetime.timedelta(milliseconds=self.data.timestamp)

    @property
    def side(self):
        """
        Returns:
            Side: the side this participant was on
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.towerType else None

    @cassiopeia.type.core.common.lazyproperty
    def victim(self):
        """
        Returns:
            Participant: the victim!
        """
        value = self.__participants[self.data.victimId] if self.data.victimId else None
        self.__count_participant()
        return value

    @property
    def ward(self):
        """
        Returns:
            Ward: the ward type associated with this event
        """
        return cassiopeia.type.core.common.Ward(self.data.wardType) if self.data.wardType else None


@cassiopeia.type.core.common.inheritdocs
class ParticipantFrame(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.ParticipantFrame

    def __init__(self, data, participants):
        super().__init__(data)
        self.__participant = participants[self.data.participantId]

    def __str__(self):
        return "Participant Frame ({player})".format(player=self.__participant)

    @property
    def gold(self):
        """
        Returns:
            int: the participant's total gold
        """
        return self.data.totalGold

    @property
    def current_gold(self):
        """
        Returns:
            int: the participant's current gold
        """
        return self.data.currentGold

    @property
    def score(self):
        """
        Returns:
            int: dominion only. the score for this participant
        """
        return self.data.dominionScore

    @property
    def jungle_monsters_killed(self):
        """
        Returns:
            int: the number of neutral jungle monsters killed
        """
        return self.data.jungleMinionsKilled

    @property
    def level(self):
        """
        Returns:
            int: the participant's champion level
        """
        return self.data.level

    @property
    def minion_kills(self):
        """
        Returns:
            int: the number of minions killed
        """
        return self.data.minionsKilled

    @property
    def participant(self):
        """
        Returns:
            Participant: the participant whose frames you are looking at
        """
        return self.__participant

    @cassiopeia.type.core.common.lazyproperty
    def position(self):
        """
        Returns:
            Position: the position of the participant
        """
        return Position(self.data.position) if self.data.position else None

    @property
    def team_score(self):
        """
        Returns:
            int: the team score for the participant
        """
        return self.data.teamScore

    @property
    def xp(self):
        """
        Returns:
            int: the amount of XP the participant has
        """
        return self.data.xp


@cassiopeia.type.core.common.inheritdocs
class Position(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.match.Position

    def __str__(self):
        return "({x}, {y})".format(x=self.x, y=self.y)

    @property
    def x(self):
        """
        Returns:
            int: the x-position of the pixel
        """
        return self.data.x

    @property
    def y(self):
        """
        Returns:
            int: the y-position of the pixel
        """
        return self.data.y


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Match.dto_type = cassiopeia.type.dto.match.MatchDetail
    Team.dto_type = cassiopeia.type.dto.match.Team
    Timeline.dto_type = cassiopeia.type.dto.match.Timeline
    ParticipantStats.dto_type = cassiopeia.type.dto.match.ParticipantStats
    ParticipantTimeline.dto_type = cassiopeia.type.dto.match.ParticipantTimeline
    Ban.dto_type = cassiopeia.type.dto.match.BannedChampion
    Frame.dto_type = cassiopeia.type.dto.match.Frame
    ParticipantTimelineData.dto_type = cassiopeia.type.dto.match.ParticipantTimelineData
    Event.dto_type = cassiopeia.type.dto.match.Event
    ParticipantFrame.dto_type = cassiopeia.type.dto.match.ParticipantFrame
    Position.dto_type = cassiopeia.type.dto.match.Position
